/*
    Copyright (C) 2004 Kor Nielsen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "oss2jack.h"

#include "fcntl.h"

#include <sys/mman.h>
#include <sys/errno.h>

#include <stdlib.h>
#include <stdio.h>

/**
 * Reallocate the buffer of OSS connection instance
 * @param instance pointer to the current oss instance
 * @param buffer_size FIXME
 */
void resize_buffer(struct circular_buffer* cbuf, int buffer_size)
{
	if(buffer_size > cbuf->buffer_size)
	{
		if(cbuf->buffer)
			munmap(cbuf->buffer, cbuf->alloc_size);
		
	cbuf->alloc_size = (buffer_size + 1) * sizeof(short);
	
	// TODO: ask for page size
	if(cbuf->alloc_size % 4096 != 0)
		cbuf->alloc_size = ((cbuf->alloc_size / 4096) + 1) * 4096;
		
		cbuf->buffer = (short*)mmap(NULL, cbuf->alloc_size, PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, 0, 0);
		if(cbuf->buffer == NULL)
		{
			perror("could not allocate memory");
			exit(-1);
		}
	}
	
	// FIXME Should buffer not be shrinked when requested size is smaller than current size?
	
	cbuf->buffer_size = buffer_size;
	cbuf->wptr = cbuf->buffer;
	cbuf->rptr = cbuf->buffer;
}

/**
 * FUSD callback method for "open" on device.
 * Here we setup our structures for the new connection
 * @param file FIXME
 * @return FIXME
 */
int do_dsp_open(struct fusd_file_info* file)
{
	pprintf(DEBUG_LEVEL_INFO, "Device opened...\n");
	struct dsp_instance* dsp = (struct dsp_instance*) malloc(sizeof(struct dsp_instance));
	struct device_info* info = (struct device_info*) file->device_info;
	struct oss_instance* instance;
	
	dsp->can_block = !(file->flags & O_NONBLOCK);
	//printf("can_block: %d\n", dsp->can_block);
	
	dsp->sample_format = AFMT_S16_NE;
	dsp->num_channels = 2;
	dsp->sample_rate = info->sample_rate;
	dsp->out_rate_converter = NULL;
	dsp->in_rate_converter = NULL;
	
	dsp->out_samples = NULL;
	dsp->out_sample_ptr = NULL;
	dsp->out_samples_left_to_mix = 0;
	dsp->out_total_samples_mixed = 0;
	dsp->polldiff = NULL;
	
	dsp->out.buffer = NULL;
	dsp->out.wptr = dsp->out.buffer;
	dsp->out.rptr = dsp->out.buffer;
	dsp->out.buffer_size = 0;
	dsp->out.dma = 0;
	dsp->out_enabled = 1;
	
	dsp->in_samples = NULL;
	dsp->in_sample_ptr = NULL;
	dsp->in_samples_left_to_read = 0;
	dsp->in_enabled = 0;
	
	dsp->in.buffer = NULL;
	dsp->in.wptr = dsp->out.buffer;
	dsp->in.rptr = dsp->out.buffer;
	dsp->in.buffer_size = 0;
	dsp->in.dma = 0;
	
	dsp->frag_size = 4096;
	dsp->frag_count = 4;
	
	dsp->out_file_info = NULL;
	dsp->in_file_info = NULL;
	
	resize_buffer(&dsp->out, dsp->frag_size * dsp->frag_count / 2);
	resize_buffer(&dsp->in, dsp->frag_size * dsp->frag_count / 2);
	
	pthread_mutex_lock(&info->mutex);
	instance = oss_find_pid(info, file->pid);
	if(instance == NULL) {
		instance = oss_create(info, file->pid, dsp, NULL);
	} else {
		instance->dsp = dsp;
	}
	pthread_mutex_unlock(&info->mutex);

	file->private_data = instance;
		
	return 0;
}

/**
 * FUSD callback method for "close" on device.
 * Here we clear our structures for the closed connection
 * @param file FIXME
 * @return FIXME
 */
int do_dsp_close(struct fusd_file_info* file)
{
	pprintf(DEBUG_LEVEL_INFO, "Device closed...\n");
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct dsp_instance* dsp = instance->dsp;
	struct device_info* info = (struct device_info*) file->device_info;
	
	pthread_mutex_lock(&info->mutex);
	
	dsp->out.wptr = dsp->out.buffer;
	dsp->out.rptr = dsp->out.buffer;
	
	instance->dsp = NULL;
	if(instance->mixer == NULL)
		oss_delete(info, instance);
	
	pthread_mutex_unlock(&info->mutex);
	
	if(dsp->out_rate_converter)
		src_delete(dsp->out_rate_converter);
	
	if(dsp->in_rate_converter)
		src_delete(dsp->in_rate_converter);
	
	
	
	if(dsp->in_sample_ptr != NULL)
	{
		dsp->in_sample_ptr = NULL;
		fusd_return(dsp->in_file_info, dsp->in_read_size);
	}
	
	if(dsp->out_sample_ptr != NULL)
	{
		dsp->out_sample_ptr = NULL;
		printf("fusd_return on out_file_info\n");
		fusd_return(dsp->out_file_info, dsp->out_write_size);
	}
	
	if(dsp->polldiff != NULL)
	{
		fusd_return(dsp->polldiff, 0);
		dsp->polldiff = 0;
	}
	
	if(dsp->in.buffer)
		munmap(dsp->in.buffer, dsp->in.alloc_size);
		
	if(dsp->out.buffer)
		munmap(dsp->out.buffer, dsp->in.alloc_size);
	
	if(dsp != NULL)
		free(dsp);
	
	return 0;
}

/**
 * FUSD callback method for "read" on device.
 * We should send out the data we got from jackd
 * TODO: This should get implemented!
 * @param file FIXME
 * @param buffer FIXME
 * @param size FIXME
 * @param offset FIXME
 * @return FIXME
 */
int do_dsp_read(struct fusd_file_info* file, char* buffer, size_t size, loff_t* offset)
{
	pprintf(DEBUG_LEVEL_DEBUG2, "Reading %d bytes from device..\n", size);
	
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct dsp_instance* dsp = instance->dsp;
	struct device_info* info = (struct device_info*) file->device_info;
	
	int sampleCount;
	if(dsp->num_channels == 2)
		sampleCount = (size >> 2) << 1;
	else
		sampleCount = (size >> 1);
	
	short* s = (short*) buffer;
	
	pthread_mutex_lock(&info->mutex);
	
	dsp->in_enabled = 1;
	
	int samples_read = cb_read_short(&dsp->in, s, sampleCount);
	
	if(!(file->flags & O_NONBLOCK) && samples_read < sampleCount)
	{
		dsp->in_samples = (short*) buffer;
		dsp->in_sample_ptr = s + samples_read;
		dsp->in_samples_left_to_read = sampleCount - samples_read;
		dsp->in_read_size = sampleCount * 2;
		dsp->in_file_info = file;
		pthread_mutex_unlock(&info->mutex);
		pprintf(DEBUG_LEVEL_DEBUG2, "Reading blocked\n");
		return -FUSD_NOREPLY;
	}
	else if((file->flags & O_NONBLOCK) && samples_read == 0)
	{
		pthread_mutex_unlock(&info->mutex);
		pprintf(DEBUG_LEVEL_DEBUG2, "Returning EAGAIN\n", samples_read * 2);
		return -EAGAIN;
	}
	
	pthread_mutex_unlock(&info->mutex);
	pprintf(DEBUG_LEVEL_DEBUG2, "Reading %d finished", samples_read * 2);
	return samples_read * 2;
} 

/**
 * FUSD callback method for "write" on device.
 * We forward the data written to the device to jackd
 * @param file FIXME
 * @param buffer FIXME
 * @param size FIXME
 * @param offset FIXME
 * @return FIXME
 */
int do_dsp_write(struct fusd_file_info *file, const char *buffer,
             size_t size, loff_t *offset)
{
	pprintf(DEBUG_LEVEL_DEBUG2, "Writing %d bytes to device: %d\n", size, (file->flags & O_NONBLOCK));
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct dsp_instance* dsp = instance->dsp;
	struct device_info* info = (struct device_info*) file->device_info;
	
	int sampleCount;
	if(dsp->num_channels == 2)
		sampleCount = (size >> 2) << 1;
	else
		sampleCount = (size >> 1);
	short* dst = (short*) buffer;
	
	pthread_mutex_lock(&info->mutex);
	
	/*
	if(cb_get_available(&dsp->out) < dsp->frag_size / 2)
	{
		if(dsp->can_block)
		{
			dsp->out_samples = (short*) buffer;
			dsp->out_sample_ptr = s;
			dsp->out_samples_left_to_mix = sampleCount;
			dsp->out_write_size = sampleCount * 2;
			dsp->out_file_info = file;
			pthread_mutex_unlock(&info->mutex);
			return -FUSD_NOREPLY;
		}
		else
		{
			pthread_mutex_unlock(&info->mutex);
			return -EAGAIN;
		}
	}*/
	
	int samples_written = cb_write_short(&dsp->out, dst, sampleCount);
	
	if(!(file->flags & O_NONBLOCK) && samples_written < sampleCount)
	{
		dsp->out_samples = (short*) buffer;
		dsp->out_sample_ptr = dst + samples_written;
		dsp->out_samples_left_to_mix = sampleCount - samples_written;
		dsp->out_write_size = sampleCount * 2;
		dsp->out_file_info = file;
		pthread_mutex_unlock(&info->mutex);
		return -FUSD_NOREPLY;
	}
	else if((file->flags & O_NONBLOCK) && samples_written == 0)
	{
		pthread_mutex_unlock(&info->mutex);
		return -EAGAIN;
	}
	
	pthread_mutex_unlock(&info->mutex);

	return samples_written * 2;
}

/**
 * FIXME
 * @param instance FIXME
 * @param dinfo FIXME
 */
void do_dsp_notify_complete(struct oss_instance* instance, struct device_info* dinfo)
{
	int curr_state, cached_state;
	struct dsp_instance* dsp = instance->dsp;
	
	if(dsp == NULL || dsp->polldiff == NULL)
	{
		return;
	}
	if(cb_get_available(&dsp->out) >= dsp->frag_size / 2)
		curr_state = FUSD_NOTIFY_OUTPUT;
	else
		curr_state = 0;
	
	if(cb_get_count(&dsp->in) > dsp->frag_size / 2)
		curr_state |= FUSD_NOTIFY_INPUT;
	
	cached_state = fusd_get_poll_diff_cached_state(dsp->polldiff);
	
	if(curr_state != cached_state)
	{
		pprintf(DEBUG_LEVEL_DEBUG2, "fusd_return\n");
		fusd_return(dsp->polldiff, curr_state);
		dsp->polldiff = NULL;
	}
}

/**
 * FUSD callback method for "polldiff" on device.
 * FIXME
 * @param file FIXME
 * @param cached_state FIXME
 * @return FIXME
 */
ssize_t do_dsp_notify_polldiff(struct fusd_file_info* file, unsigned int cached_state)
{
	pprintf(DEBUG_LEVEL_DEBUG2, "PollDiff to device...\n");
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct dsp_instance* dsp = instance->dsp;
	struct device_info* info = (struct device_info*) file->device_info;
	
	if(dsp == NULL)
		return -EINVAL;
	
	pthread_mutex_lock(&info->mutex);
	
	if(dsp->polldiff != NULL)
	{
		fusd_destroy(dsp->polldiff);
		dsp->polldiff = NULL;
	}
	
	dsp->polldiff = file;
	do_dsp_notify_complete(instance, info);
	
	pthread_mutex_unlock(&info->mutex);
	
	return -FUSD_NOREPLY;
}


int do_dsp_mmap(struct fusd_file_info *file, int offset, size_t length, int flags, void** addr, size_t* out_length)
{
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct dsp_instance* dsp = instance->dsp;
	*addr = dsp->out.buffer;
	*out_length = dsp->out.alloc_size;
	
	dsp->out.dma = 1;
	return 0;
}

/**
 * FUSD callback method for "ioctl" on device.
 * Here we process the OSS IOCTLs
 * @param file FIXME
 * @param request FIXME
 * @param data FIXME
 * @return FIXME
 */
int do_dsp_ioctl(struct fusd_file_info *file, int request, void *data)
{
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct dsp_instance* dsp = (struct dsp_instance*) instance->dsp;
	struct device_info* dinfo = (struct device_info*) file->device_info;
	
	pthread_mutex_lock(&dinfo->mutex);
	int return_value = -1;
	switch(request)
	{
		case SNDCTL_DSP_RESET:
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_RESET\n");
			dsp->out.wptr = dsp->out.buffer;
			dsp->out.rptr = dsp->out.buffer;
			//dsp->samples_mixed = dinfo->samples_sent + 64 * 2;
			//dsp->wptr = g_wptr + 64 * 2;
			//if(dsp->wptr > g_wend)
			//	dsp->wptr -= dsp->buffer_size;
			return_value = 0;
			break;
		
		case SNDCTL_DSP_SETFMT:
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_SETFMT\n");
			pprintf(DEBUG_LEVEL_DEBUG1, "Preferred format: %d\n", *(int*)data);
			*(int*)data = AFMT_S16_NE;
			return_value = 0;
			break;
		
		case SNDCTL_DSP_GETFMTS:
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_GETFMTS\n");
			*(int*)data = AFMT_S16_NE;
			return_value = 0;
			break;
		
		case SNDCTL_DSP_CHANNELS:
			{
				pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_CHANNELS  (%d)\n", *(int*)data);
				int nChannels = *(int*)data;
				if(nChannels <= 0)
					*(int*)data = 1;
				else if(nChannels > 2)
					*(int*)data = 2;
				
				dsp->num_channels = nChannels;
				
				*(int*)data = dsp->num_channels;
				pprintf(DEBUG_LEVEL_DEBUG2,"    num_channels: %d\n", dsp->num_channels);
				
				if(dsp->out_rate_converter != NULL)
				{
					int err;
					src_delete(dsp->out_rate_converter);
					dsp->out_rate_converter = src_new(SRC_LINEAR, dsp->num_channels, &err);
				}
				if(dsp->in_rate_converter != NULL)
				{
					int err;
					src_delete(dsp->in_rate_converter);
					dsp->in_rate_converter = src_new(SRC_LINEAR, dsp->num_channels, &err);
				}
			}
			return_value = 0;
			break;
			
		case SNDCTL_DSP_STEREO:
			{
				pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_STEREO  (%d)\n", *(int*)data);
				int nChannels = *(int*)data + 1;
				if(nChannels <= 0)
					nChannels = 1;
				else if(nChannels > 2)
					nChannels = 2;
				
				dsp->num_channels = nChannels;
				
				if(dsp->out_rate_converter != NULL)
				{
					int err;
					src_delete(dsp->out_rate_converter);
					dsp->out_rate_converter = src_new(SRC_LINEAR, dsp->num_channels, &err);
				}
				if(dsp->in_rate_converter != NULL)
				{
					int err;
					src_delete(dsp->in_rate_converter);
					dsp->in_rate_converter = src_new(SRC_LINEAR, dsp->num_channels, &err);
				}
			}
			return_value = 0;
			break;
		
		case SNDCTL_DSP_SPEED:
			{
				pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_SPEED  (%d)\n", *(int*)data);
				int err = 0;
				int rate = *(int*)data;
				if(rate < 8000)
					rate = 44100;

				if(dsp->out_rate_converter == NULL && (*(int*)data) != dinfo->sample_rate)
				{
					dsp->out_rate_converter = src_new(SRC_LINEAR, dsp->num_channels, &err);
					if(err)
					{
						*(int*)data = dsp->sample_rate;
						return_value = 0;
					}
				}
				
				if(dsp->in_rate_converter == NULL && (*(int*)data) != dinfo->sample_rate)
				{
					dsp->in_rate_converter = src_new(SRC_LINEAR, dsp->num_channels, &err);
					if(err)
					{
						*(int*)data = dsp->sample_rate;
						return_value = 0;
					}
				}
				
				dsp->out_rate_ratio = (double)dinfo->sample_rate / (double)rate;
				dsp->in_rate_ratio = (double)rate / (double)dinfo->sample_rate;
				
				dsp->sample_rate = rate;
				*(int*)data = rate;
			}
			return_value = 0;
			break;
			
		case SNDCTL_DSP_SYNC:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_SYNC  NOT IMPLEMENTED\n");
			return_value = 0;
			break;
		
		
		case SNDCTL_DSP_GETBLKSIZE:
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_GETBLKSIZE\n");
			*((int*)data) = (dsp->frag_size) * 2;
			return_value = 0;
			break;
		
		case SOUND_PCM_WRITE_FILTER:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SOUND_PCM_WRITE_FILTER  NOT IMPLEMENTED\n");
			break;
		
		case SNDCTL_DSP_POST:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_POST  NOT IMPLEMENTED\n");
			break;
			
		case SNDCTL_DSP_SUBDIVIDE:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_SUBDIVIDE  NOT IMPLEMENTED\n");
			break;
		
		case SNDCTL_DSP_SETFRAGMENT:
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_SETFRAGMENT  (%d)\n", (*(int*)data) & 0xffff);
			dsp->frag_size = (1 << ((*(int*)data) & 0xffff));
			dsp->frag_count = ((*(int*)data) & 0xffff0000) >> 16;
			
			if(dsp->frag_count > 8)
				dsp->frag_count = 8;
			
			(*(int*)data) = ((*(int*)data) & 0xffff ) | dsp->frag_count << 16;
			
			pprintf(DEBUG_LEVEL_DEBUG1, "    FragSize: %d x %d\n", dsp->frag_size, dsp->frag_count);
			resize_buffer(&dsp->out, dsp->frag_size * dsp->frag_count / 2);
			resize_buffer(&dsp->in, dsp->frag_size * dsp->frag_count / 2);
			return_value = 0;
			break;
		
		case SNDCTL_DSP_GETOSPACE:
		{
			pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SNDCTL_DSP_GETOSPACE\n");
			struct audio_buf_info* info = (struct audio_buf_info*) data;
			info->fragments = cb_get_available(&dsp->out) / (dsp->out.buffer_size / dsp->frag_count);
			if(info->fragments > dsp->frag_count)
				info->fragments = dsp->frag_count;
			info->fragstotal = dsp->frag_count;
			info->fragsize = (dsp->out.buffer_size / dsp->frag_count) * 2;
			info->bytes = (dsp->out.buffer_size - cb_get_count(&dsp->out)) * 2;
			pprintf(DEBUG_LEVEL_DEBUG2, "    Fragments: %d\n    Fragstotal: %d\n    Fragsize: %d\n    Bytes: %d\n", info->fragments, info->fragstotal, info->fragsize, info->bytes);
			return_value = 0;
			break;
		}

		case SNDCTL_DSP_GETISPACE:
			pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SNDCTL_DSP_GETISPACE\n");
			struct audio_buf_info* info = (struct audio_buf_info*) data;
			
			info->fragments = cb_get_count(&dsp->in) / (dsp->in.buffer_size / dsp->frag_count);
			if(info->fragments > dsp->frag_count)
				info->fragments = dsp->frag_count;
			info->fragstotal = dsp->frag_count;
			info->fragsize = (dsp->in.buffer_size / dsp->frag_count) * 2;
			info->bytes = cb_get_count(&dsp->in) * 2;
			pprintf(DEBUG_LEVEL_DEBUG2, "    Fragstotal: %d\n    Fragsize: %d\n    Bytes: %d\n", info->fragstotal, info->fragsize, info->bytes);
			return_value = 0;
			break;
		
		case SNDCTL_DSP_NONBLOCK:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_NONBLOCK  NOT IMPLEMENTED\n");
			break;
		
		case SNDCTL_DSP_GETCAPS:
			*(int*)data = DSP_CAP_TRIGGER | DSP_CAP_MMAP | DSP_CAP_DUPLEX;
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_GETCAPS (%d)\n", *(int*)data);
			return_value = 0;
			break;
			
		case SNDCTL_DSP_SETDUPLEX:
			dsp->in_enabled = 1;
			return_value = 0;
			break;
		
		case SNDCTL_DSP_GETTRIGGER:
			*(int*)data = PCM_ENABLE_OUTPUT;
			if(dsp->in_enabled)
				*(int*)data |= PCM_ENABLE_INPUT;
			return_value = 0;
			break;
		
		case SNDCTL_DSP_SETTRIGGER:
			pprintf(DEBUG_LEVEL_DEBUG1, "IOCTL: SNDCTL_DSP_SETTRIGGER\n");
			if((*(int*)data) & PCM_ENABLE_OUTPUT)
			{
				dsp->out_enabled = 1;
			}
			else
			{
				dsp->out_enabled = 0;
			}
			return_value = 0;
			break;
			
		case SNDCTL_DSP_GETIPTR:
			pprintf(DEBUG_LEVEL_DEBUG2,  "IOCTL: SNDCTL_DSP_GETIPTR\n");
			
			((count_info*)data)->bytes = dsp->in_total_samples_read * 2;
			((count_info*)data)->blocks = 0;
			((count_info*)data)->ptr = 0;
			
			return_value = 0;
			break;
			
		case SNDCTL_DSP_GETOPTR:
			pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SNDCTL_DSP_GETOPTR\n");
			((count_info*)data)->bytes = dsp->out_total_samples_mixed * 2 + dinfo->last_frame_size * dsp->num_channels * 2;
			
			((count_info*)data)->blocks = dsp->out_total_samples_mixed * 2 / dsp->frag_size;
			((count_info*)data)->ptr = (dsp->out.rptr - dsp->out.buffer) * 2;
			
			return_value = 0;
			break;

		case SNDCTL_DSP_MAPINBUF:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_MAPINBUF  NOT IMPLEMENTED\n");

			break;
			
		case SNDCTL_DSP_MAPOUTBUF:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_MAPOUTBUF  NOT IMPLEMENTED\n");
			break;
			
		case SNDCTL_DSP_SETSYNCRO:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_SETSYNCRO  NOT IMPLEMENTED\n");
			break;
			
		case SNDCTL_DSP_GETODELAY:
			*((int*)data) = cb_get_count(&dsp->out) * 2 + dinfo->last_frame_size * dsp->num_channels * 2;
			return_value = 0;
			pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SNDCTL_DSP_GETODELAY bytes=%d\n", *((int*)data));
			break;
			
		case SNDCTL_DSP_GETCHANNELMASK:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_GETCHANNELMASK  NOT IMPLEMENTED\n");
			break;
			
		case SNDCTL_DSP_BIND_CHANNEL:
			pprintf(DEBUG_LEVEL_NOTICE, "IOCTL: SNDCTL_DSP_BIND_CHANNEL  NOT IMPLEMENTED\n");
			break;
		
		default:
			pprintf(DEBUG_LEVEL_WARNING, "IOCTL: Unknown IOCTL %d,  NOT IMPLEMENTED\n", request);
			break;
	}
	
	pthread_mutex_unlock(&dinfo->mutex);
	
	return return_value;
}

struct fusd_file_operations dsp_file_ops = 
{
	open: do_dsp_open,
	read: do_dsp_read,
	write: do_dsp_write,
	close: do_dsp_close,
	ioctl: do_dsp_ioctl,
	mmap: do_dsp_mmap,
	poll_diff: do_dsp_notify_polldiff
};
